from ._core import (
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    GraphQLList,
    ResourcesWrite,
    ResourcesWriteResult,
    PageInfo,
)
from ._alert import (
    Alert,
    AlertApply,
    AlertApplyList,
    AlertFields,
    AlertGraphQL,
    AlertList,
    AlertTextFields,
    AlertWrite,
    AlertWriteList,
)
from ._basic_bid_matrix import (
    BasicBidMatrix,
    BasicBidMatrixApply,
    BasicBidMatrixApplyList,
    BasicBidMatrixFields,
    BasicBidMatrixGraphQL,
    BasicBidMatrixList,
    BasicBidMatrixTextFields,
    BasicBidMatrixWrite,
    BasicBidMatrixWriteList,
)
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentGraphQL,
    BidDocumentList,
    BidDocumentTextFields,
    BidDocumentWrite,
    BidDocumentWriteList,
)
from ._bid_matrix import (
    BidMatrix,
    BidMatrixApply,
    BidMatrixApplyList,
    BidMatrixFields,
    BidMatrixGraphQL,
    BidMatrixList,
    BidMatrixTextFields,
    BidMatrixWrite,
    BidMatrixWriteList,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodGraphQL,
    BidMethodList,
    BidMethodTextFields,
    BidMethodWrite,
    BidMethodWriteList,
)
from ._multi_scenario_matrix import (
    MultiScenarioMatrix,
    MultiScenarioMatrixApply,
    MultiScenarioMatrixApplyList,
    MultiScenarioMatrixFields,
    MultiScenarioMatrixGraphQL,
    MultiScenarioMatrixList,
    MultiScenarioMatrixTextFields,
    MultiScenarioMatrixWrite,
    MultiScenarioMatrixWriteList,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaGraphQL,
    PriceAreaList,
    PriceAreaTextFields,
    PriceAreaWrite,
    PriceAreaWriteList,
)
from ._shop_multi_scenario_method import (
    SHOPMultiScenarioMethod,
    SHOPMultiScenarioMethodApply,
    SHOPMultiScenarioMethodApplyList,
    SHOPMultiScenarioMethodFields,
    SHOPMultiScenarioMethodGraphQL,
    SHOPMultiScenarioMethodList,
    SHOPMultiScenarioMethodTextFields,
    SHOPMultiScenarioMethodWrite,
    SHOPMultiScenarioMethodWriteList,
)
from ._shop_price_scenario import (
    SHOPPriceScenario,
    SHOPPriceScenarioApply,
    SHOPPriceScenarioApplyList,
    SHOPPriceScenarioFields,
    SHOPPriceScenarioGraphQL,
    SHOPPriceScenarioList,
    SHOPPriceScenarioTextFields,
    SHOPPriceScenarioWrite,
    SHOPPriceScenarioWriteList,
)
from ._shop_price_scenario_result import (
    SHOPPriceScenarioResult,
    SHOPPriceScenarioResultApply,
    SHOPPriceScenarioResultApplyList,
    SHOPPriceScenarioResultFields,
    SHOPPriceScenarioResultGraphQL,
    SHOPPriceScenarioResultList,
    SHOPPriceScenarioResultWrite,
    SHOPPriceScenarioResultWriteList,
)
from ._water_value_based_method import (
    WaterValueBasedMethod,
    WaterValueBasedMethodApply,
    WaterValueBasedMethodApplyList,
    WaterValueBasedMethodFields,
    WaterValueBasedMethodGraphQL,
    WaterValueBasedMethodList,
    WaterValueBasedMethodTextFields,
    WaterValueBasedMethodWrite,
    WaterValueBasedMethodWriteList,
)

BasicBidMatrix.model_rebuild()
BasicBidMatrixGraphQL.model_rebuild()
BasicBidMatrixWrite.model_rebuild()
BasicBidMatrixApply.model_rebuild()
BidDocument.model_rebuild()
BidDocumentGraphQL.model_rebuild()
BidDocumentWrite.model_rebuild()
BidDocumentApply.model_rebuild()
BidMatrix.model_rebuild()
BidMatrixGraphQL.model_rebuild()
BidMatrixWrite.model_rebuild()
BidMatrixApply.model_rebuild()
MultiScenarioMatrix.model_rebuild()
MultiScenarioMatrixGraphQL.model_rebuild()
MultiScenarioMatrixWrite.model_rebuild()
MultiScenarioMatrixApply.model_rebuild()
PriceArea.model_rebuild()
PriceAreaGraphQL.model_rebuild()
PriceAreaWrite.model_rebuild()
PriceAreaApply.model_rebuild()
SHOPMultiScenarioMethod.model_rebuild()
SHOPMultiScenarioMethodGraphQL.model_rebuild()
SHOPMultiScenarioMethodWrite.model_rebuild()
SHOPMultiScenarioMethodApply.model_rebuild()
SHOPPriceScenarioResult.model_rebuild()
SHOPPriceScenarioResultGraphQL.model_rebuild()
SHOPPriceScenarioResultWrite.model_rebuild()
SHOPPriceScenarioResultApply.model_rebuild()


__all__ = [
    "DataRecord",
    "DataRecordGraphQL",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "GraphQLCore",
    "GraphQLList",
    "ResourcesWriteResult",
    "PageInfo",
    "Alert",
    "AlertGraphQL",
    "AlertWrite",
    "AlertApply",
    "AlertList",
    "AlertWriteList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BasicBidMatrix",
    "BasicBidMatrixGraphQL",
    "BasicBidMatrixWrite",
    "BasicBidMatrixApply",
    "BasicBidMatrixList",
    "BasicBidMatrixWriteList",
    "BasicBidMatrixApplyList",
    "BasicBidMatrixFields",
    "BasicBidMatrixTextFields",
    "BidDocument",
    "BidDocumentGraphQL",
    "BidDocumentWrite",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidMatrix",
    "BidMatrixGraphQL",
    "BidMatrixWrite",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMethod",
    "BidMethodGraphQL",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "MultiScenarioMatrix",
    "MultiScenarioMatrixGraphQL",
    "MultiScenarioMatrixWrite",
    "MultiScenarioMatrixApply",
    "MultiScenarioMatrixList",
    "MultiScenarioMatrixWriteList",
    "MultiScenarioMatrixApplyList",
    "MultiScenarioMatrixFields",
    "MultiScenarioMatrixTextFields",
    "PriceArea",
    "PriceAreaGraphQL",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "SHOPMultiScenarioMethod",
    "SHOPMultiScenarioMethodGraphQL",
    "SHOPMultiScenarioMethodWrite",
    "SHOPMultiScenarioMethodApply",
    "SHOPMultiScenarioMethodList",
    "SHOPMultiScenarioMethodWriteList",
    "SHOPMultiScenarioMethodApplyList",
    "SHOPMultiScenarioMethodFields",
    "SHOPMultiScenarioMethodTextFields",
    "SHOPPriceScenario",
    "SHOPPriceScenarioGraphQL",
    "SHOPPriceScenarioWrite",
    "SHOPPriceScenarioApply",
    "SHOPPriceScenarioList",
    "SHOPPriceScenarioWriteList",
    "SHOPPriceScenarioApplyList",
    "SHOPPriceScenarioFields",
    "SHOPPriceScenarioTextFields",
    "SHOPPriceScenarioResult",
    "SHOPPriceScenarioResultGraphQL",
    "SHOPPriceScenarioResultWrite",
    "SHOPPriceScenarioResultApply",
    "SHOPPriceScenarioResultList",
    "SHOPPriceScenarioResultWriteList",
    "SHOPPriceScenarioResultApplyList",
    "SHOPPriceScenarioResultFields",
    "WaterValueBasedMethod",
    "WaterValueBasedMethodGraphQL",
    "WaterValueBasedMethodWrite",
    "WaterValueBasedMethodApply",
    "WaterValueBasedMethodList",
    "WaterValueBasedMethodWriteList",
    "WaterValueBasedMethodApplyList",
    "WaterValueBasedMethodFields",
    "WaterValueBasedMethodTextFields",
]
